@@ -16,9 +16,10 @@ use ptr_::Unique;
 use core::{cmp, fmt, mem, usize, ops::DerefMut, num::NonZeroUsize,
            ptr::{self, NonNull}};
 
-macro_rules! try_opt {
-    ($x:expr) => (match $x { None => return None, Some(x) => x })
-}
+/// Represents the combination of a starting address and
+/// a total capacity of the returned block.
+#[derive(Debug)]
+pub struct Excess(pub NonNull<u8>, pub usize);
 
 /// Layout of a block of memory.
 ///
@@ -48,6 +49,7 @@ pub struct Layout {
     align: NonZeroUsize,
 }
 
+
 // FIXME: audit default implementations for overflow errors,
 // (potentially switching to overflowing_add and
 //  overflowing_mul as necessary).
@@ -65,7 +67,7 @@ impl Layout {
     ///    must not overflow (i.e. the rounded value must be less than
     ///    `usize::MAX`).
     #[inline]
-    pub const fn from_size_align(size: usize, align: usize) -> Option<Layout> {
+    pub fn from_size_align(size: usize, align: usize) -> Option<Layout> {
         if !align.is_power_of_two() {
             return None;
         }
@@ -110,15 +112,15 @@ impl Layout {
 
     /// The minimum size in bytes for a memory block of this layout.
     #[inline]
-    pub const fn size(&self) -> usize { self.size }
+    pub fn size(&self) -> usize { self.size }
 
     /// The minimum byte alignment for a memory block of this layout.
     #[inline]
-    pub const fn align(&self) -> NonZeroUsize { self.align }
+    pub fn align(&self) -> NonZeroUsize { self.align }
 
     /// Constructs a `Layout` suitable for holding a value of type `T`.
     #[inline]
-    pub const fn new<T>() -> Self {
+    pub fn new<T>() -> Self {
         Layout { size: mem::size_of::<T>(),
                  align: unsafe { NonZeroUsize::new_unchecked(mem::align_of::<T>()) } }
     }
@@ -149,8 +151,8 @@ impl Layout {
     /// Panics if the combination of `self.size` and the given `align`
     /// violates the conditions listed in `from_size_align`.
     #[inline]
-    pub const fn align_to(&self, align: usize) -> Self {
-        Layout { size: self.size, align: if align > self.align.get() { unsafe { NonZeroUsize::new_unchecked(align) } } else { self.align } }
+    pub fn align_to(&self, align: usize) -> Self {
+        Layout::from_size_align(self.size, cmp::max(self.align.get(), align)).unwrap()
     }
 
     /// Returns the amount of padding we must insert after `self`
@@ -170,7 +172,7 @@ impl Layout {
     /// address for the whole allocated block of memory. One way to
     /// satisfy this constraint is to ensure `align <= self.align`.
     #[inline]
-    pub const fn padding_needed_for(&self, align: NonZeroUsize) -> usize {
+    pub fn padding_needed_for(&self, align: NonZeroUsize) -> usize {
         let len = self.size();
         let align = align.get();
 
@@ -210,7 +212,7 @@ impl Layout {
     /// address for the whole allocated block of memory. One way to
     /// satisfy this constraint is to ensure `align <= self.align`.
     #[inline]
-    pub const fn pad_to(&self, align: NonZeroUsize) -> Self {
+    pub fn pad_to(&self, align: NonZeroUsize) -> Self {
         Layout {
             size: {
                 let align = align.get();
@@ -229,15 +231,15 @@ impl Layout {
     ///
     /// On arithmetic overflow, returns `None`.
     #[inline]
-    pub const fn repeat(&self, n: usize) -> Option<(Self, usize)> {
-        let padded_size = try_opt!(self.size.checked_add(self.padding_needed_for(self.align)));
-        let alloc_size = try_opt!(padded_size.checked_mul(n));
+    pub fn repeat(&self, n: usize) -> Option<(Self, usize)> {
+        let padded_size = self.size.checked_add(self.padding_needed_for(self.align))?;
+        let alloc_size = padded_size.checked_mul(n)?;
 
         // We can assume that `self.align` is a power-of-two that does
         // not exceed 2^31. Furthermore, `alloc_size` has already been
         // rounded up to a multiple of `self.align`; therefore, the
         // call to `Layout::from_size_align` below should never panic.
-        Some((Layout { size: alloc_size, align: self.align }, padded_size))
+        Some((Layout::from_size_align(alloc_size, self.align.get()).unwrap(), padded_size))
     }
 
     /// Creates a layout describing the record for `self` followed by
@@ -252,15 +254,15 @@ impl Layout {
     ///
     /// On arithmetic overflow, returns `None`.
     #[inline]
-    pub const fn extend(&self, next: Self) -> Option<(Self, usize)> {
-        let new_align = if self.align.get() > next.align.get() { self.align } else { next.align };
-        let realigned = try_opt!(Layout::from_size_align(self.size, new_align.get()));
+    pub fn extend(&self, next: Self) -> Option<(Self, usize)> {
+        let new_align = cmp::max(self.align, next.align);
+        let realigned = Layout::from_size_align(self.size, new_align.get())?;
 
         let pad = realigned.padding_needed_for(next.align);
 
-        let offset = try_opt!(self.size.checked_add(pad));
-        let new_size = try_opt!(offset.checked_add(next.size));
-        Some((try_opt!(Layout::from_size_align(new_size, new_align.get())), offset))
+        let offset = self.size.checked_add(pad)?;
+        let new_size = offset.checked_add(next.size)?;
+        Some((Layout::from_size_align(new_size, new_align.get())?, offset))
     }
 
     /// Creates a layout describing the record for `n` instances of
@@ -276,8 +278,8 @@ impl Layout {
     ///
     /// On arithmetic overflow, returns `None`.
     #[inline]
-    pub const fn repeat_packed(&self, n: usize) -> Option<Self> {
-        Layout::from_size_align(try_opt!(self.size().checked_mul(n)), self.align.get())
+    pub fn repeat_packed(&self, n: usize) -> Option<Self> {
+        Layout::from_size_align(self.size().checked_mul(n)?, self.align.get())
     }
 
     /// Creates a layout describing the record for `self` followed by
@@ -296,19 +298,20 @@ impl Layout {
     ///
     /// On arithmetic overflow, returns `None`.
     #[inline]
-    pub const fn extend_packed(&self, next: Self) -> Option<(Self, usize)> {
-        let new_size = try_opt!(self.size().checked_add(next.size()));
-        Some((try_opt!(Layout::from_size_align(new_size, self.align.get())), self.size()))
+    pub fn extend_packed(&self, next: Self) -> Option<(Self, usize)> {
+        let new_size = self.size().checked_add(next.size())?;
+        Some((Layout::from_size_align(new_size, self.align.get())?, self.size()))
     }
 
     /// Creates a layout describing the record for a `[T; n]`.
     ///
     /// On arithmetic overflow, returns `None`.
     #[inline]
-    pub const fn array<T>(n: usize) -> Option<Self> {
-        let (k, _offs) = try_opt!(Layout::new::<T>().repeat(n));
-        //debug_assert!(offs == mem::size_of::<T>());
-        Some(k)
+    pub fn array<T>(n: usize) -> Option<Self> {
+        Layout::new::<T>().repeat(n).map(|(k, offs)| {
+            debug_assert!(offs == mem::size_of::<T>());
+            k
+        })
     }
 }
 
@@ -337,19 +340,19 @@ pub enum AllocErr {
 
 impl AllocErr {
     #[inline]
-    pub const fn invalid_input(details: &'static str) -> Self {
+    pub fn invalid_input(details: &'static str) -> Self {
         AllocErr::Unsupported { details }
     }
     #[inline]
-    pub const fn is_memory_exhausted(&self) -> bool {
+    pub fn is_memory_exhausted(&self) -> bool {
         if let AllocErr::Exhausted { .. } = *self { true } else { false }
     }
     #[inline]
-    pub const fn is_request_unsupported(&self) -> bool {
+    pub fn is_request_unsupported(&self) -> bool {
         if let AllocErr::Unsupported { .. } = *self { true } else { false }
     }
     #[inline]
-    pub const fn description(&self) -> &str {
+    pub fn description(&self) -> &str {
         match *self {
             AllocErr::Exhausted { .. } => "allocator memory exhausted",
             AllocErr::Unsupported { .. } => "unsupported allocator request",
@@ -377,12 +380,6 @@ impl fmt::Display for CannotReallocInPlace {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result { write!(f, "{}", self.description()) }
 }
 
-#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
-pub struct Allocation {
-    pub ptr: NonNull<u8>,
-    pub size: usize,
-}
-
 /// An implementation of `Alloc` can allocate, reallocate, and
 /// deallocate arbitrary blocks of data described via `Layout`.
 ///
@@ -391,7 +388,8 @@ pub struct Allocation {
 ///
 /// * the starting address for that memory block was previously
 ///   returned by a previous call to an allocation method (`alloc`,
-///   `alloc_one`, `alloc_array`) or reallocation method (`realloc`,
+///   `alloc_zeroed`, `alloc_excess`, `alloc_one`, `alloc_array`) or
+///   reallocation method (`realloc`, `realloc_excess`, or
 ///   `realloc_array`), and
 ///
 /// * the memory block has not been subsequently deallocated, where
@@ -512,7 +510,7 @@ pub unsafe trait Alloc {
     /// Clients wishing to abort computation in response to an
     /// allocation error are encouraged to call the allocator's `oom`
     /// method, rather than directly invoking `panic!` or similar.
-    unsafe fn alloc(&mut self, layout: Layout) -> Result<Allocation, AllocErr>;
+    unsafe fn alloc(&mut self, layout: Layout) -> Result<NonNull<u8>, AllocErr>;
 
     /// Deallocate the memory referenced by `ptr`.
     ///
@@ -564,7 +562,7 @@ pub unsafe trait Alloc {
     fn usable_size(&self, layout: Layout) -> (usize, usize) { (layout.size(), layout.size()) }
 
     // == METHODS FOR MEMORY REUSE ==
-    // realloc.
+    // realloc. alloc_excess, realloc_excess
 
     /// Returns a pointer suitable for holding data described by
     /// `new_layout`, meeting its size and alignment guarantees. To
@@ -632,21 +630,16 @@ pub unsafe trait Alloc {
     unsafe fn realloc(&mut self,
                       ptr: NonNull<u8>,
                       layout: Layout,
-                      new_size: usize,
-                      in_place: bool) -> Result<Allocation, AllocErr> {
+                      new_size: usize) -> Result<NonNull<u8>, AllocErr> {
         let old_size = layout.size();
 
         if let Ok(()) = self.resize_in_place(ptr, layout, new_size) {
-            return Ok(Allocation { ptr, size: new_size });
-        }
-
-        if in_place {
-            return Err(AllocErr::Exhausted { request: layout });
+            return Ok(ptr);
         }
 
         // otherwise, fall back on alloc + copy + dealloc.
         let result = self.alloc(Layout { size: new_size, align: layout.align() });
-        if let Ok(Allocation { ptr: new_ptr, .. }) = result {
+        if let Ok(new_ptr) = result {
             ptr::copy_nonoverlapping(ptr.as_ptr(), new_ptr.as_ptr(),
                                      cmp::min(old_size, new_size));
             self.dealloc(ptr, layout);
@@ -654,6 +647,82 @@ pub unsafe trait Alloc {
         result
     }
 
+    /// Behaves like `alloc`, but also ensures that the contents
+    /// are set to zero before being returned.
+    ///
+    /// # Safety
+    ///
+    /// This function is unsafe for the same reasons that `alloc` is.
+    ///
+    /// # Errors
+    ///
+    /// Returning `Err` indicates that either memory is exhausted or
+    /// `layout` does not meet allocator's size or alignment
+    /// constraints, just as in `alloc`.
+    ///
+    /// Clients wishing to abort computation in response to an
+    /// allocation error are encouraged to call the allocator's `oom`
+    /// method, rather than directly invoking `panic!` or similar.
+    #[inline]
+    unsafe fn alloc_zeroed(&mut self, layout: Layout) -> Result<NonNull<u8>, AllocErr> {
+        let size = layout.size();
+        let r = self.alloc(layout);
+        if let Ok(p) = r {
+            ptr::write_bytes(p.as_ptr(), 0, size);
+        }
+        r
+    }
+
+    /// Behaves like `alloc`, but also returns the whole size of
+    /// the returned block. For some `layout` inputs, like arrays, this
+    /// may include extra storage usable for additional data.
+    ///
+    /// # Safety
+    ///
+    /// This function is unsafe for the same reasons that `alloc` is.
+    ///
+    /// # Errors
+    ///
+    /// Returning `Err` indicates that either memory is exhausted or
+    /// `layout` does not meet allocator's size or alignment
+    /// constraints, just as in `alloc`.
+    ///
+    /// Clients wishing to abort computation in response to an
+    /// allocation error are encouraged to call the allocator's `oom`
+    /// method, rather than directly invoking `panic!` or similar.
+    #[inline]
+    unsafe fn alloc_excess(&mut self, layout: Layout) -> Result<Excess, AllocErr> {
+        let usable_size = self.usable_size(layout);
+        self.alloc(layout).map(|p| Excess(p, usable_size.1))
+    }
+
+    /// Behaves like `realloc`, but also returns the whole size of
+    /// the returned block. For some `layout` inputs, like arrays, this
+    /// may include extra storage usable for additional data.
+    ///
+    /// # Safety
+    ///
+    /// This function is unsafe for the same reasons that `realloc` is.
+    ///
+    /// # Errors
+    ///
+    /// Returning `Err` indicates that either memory is exhausted or
+    /// `layout` does not meet allocator's size or alignment
+    /// constraints, just as in `realloc`.
+    ///
+    /// Clients wishing to abort computation in response to an
+    /// reallocation error are encouraged to call the allocator's `oom`
+    /// method, rather than directly invoking `panic!` or similar.
+    #[inline]
+    unsafe fn realloc_excess(&mut self,
+                             ptr: NonNull<u8>,
+                             layout: Layout,
+                             new_size: usize) -> Result<Excess, AllocErr> {
+        let new_layout = Layout { size: new_size, align: layout.align() };
+        let usable_size = self.usable_size(new_layout);
+        self.realloc(ptr, layout, new_size).map(|p| Excess(p, usable_size.1))
+    }
+
     /// Attempts to resize the allocation referenced by `ptr` to fit `new_layout`.
     ///
     /// If this returns `Ok`, then the allocator has asserted that the
@@ -742,7 +811,7 @@ pub unsafe trait Alloc {
     fn alloc_one<T>(&mut self) -> Result<Unique<T>, AllocErr> {
         let k = Layout::new::<T>();
         if k.size() > 0 {
-            unsafe { self.alloc(k).map(|Allocation { ptr: p, .. }| p.cast().into()) }
+            unsafe { self.alloc(k).map(|p| p.cast().into()) }
         } else {
             Err(AllocErr::invalid_input("zero-sized type invalid for alloc_one"))
         }
@@ -807,8 +876,8 @@ pub unsafe trait Alloc {
     fn alloc_array<T>(&mut self, n: usize) -> Result<(Unique<T>, usize), AllocErr> {
         match Layout::array::<T>(n) {
             Some(ref layout) if layout.size() > 0 => { unsafe {
-                self.alloc(layout.clone())
-                    .map(|Allocation { ptr: p, size: n }| (p.cast().into(), n / mem::size_of::<T>()))
+                self.alloc_excess(layout.clone())
+                    .map(|Excess(p, n)| (p.cast().into(), n / mem::size_of::<T>()))
             } },
             _ => Err(AllocErr::invalid_input("invalid layout for alloc_array")),
         }
@@ -855,8 +924,8 @@ pub unsafe trait Alloc {
                                n_new: usize) -> Result<(Unique<T>, usize), AllocErr> {
         match (Layout::array::<T>(n_old), Layout::array::<T>(n_new), ptr.as_ptr()) {
             (Some(ref k_old), Some(ref k_new), ptr) if k_old.size() > 0 && k_new.size() > 0 => {
-                self.realloc(ptr.cast(), k_old.clone(), k_new.size(), false)
-                    .map(|Allocation { ptr: p, size: n }| (p.cast().into(),
+                self.realloc_excess(ptr.cast(), k_old.clone(), k_new.size())
+                    .map(|Excess(p, n)| (p.cast().into(),
                                          n / mem::size_of::<T>()))
             }
             _ => {
@@ -902,14 +971,17 @@ pub unsafe trait Alloc {
 pub struct NullAllocator(());
 
 unsafe impl Alloc for NullAllocator {
-    #[inline] unsafe fn alloc(&mut self, _: Layout) -> Result<Allocation, AllocErr> { Err(AllocErr::Unsupported { details: "" }) }
+    #[inline] unsafe fn alloc(&mut self, _: Layout) -> Result<NonNull<u8>, AllocErr> { Err(AllocErr::Unsupported { details: "" }) }
     #[inline] unsafe fn dealloc(&mut self, _: NonNull<u8>, _: Layout) {}
 }
 
 unsafe impl<A: Alloc + ?Sized, P: DerefMut<Target = A>> Alloc for P {
-    #[inline] unsafe fn alloc(&mut self, l: Layout) -> Result<Allocation, AllocErr> { self.deref_mut().alloc(l) }
+    #[inline] unsafe fn alloc(&mut self, l: Layout) -> Result<NonNull<u8>, AllocErr> { self.deref_mut().alloc(l) }
     #[inline] unsafe fn dealloc(&mut self, ptr: NonNull<u8>, l: Layout) { self.deref_mut().dealloc(ptr, l) }
-    #[inline] unsafe fn realloc(&mut self, ptr: NonNull<u8>, old_l: Layout, new_size: usize, in_place: bool) -> Result<Allocation, AllocErr> { self.deref_mut().realloc(ptr, old_l, new_size, in_place) }
+    #[inline] unsafe fn realloc(&mut self, ptr: NonNull<u8>, old_l: Layout, new_size: usize) -> Result<NonNull<u8>, AllocErr> { self.deref_mut().realloc(ptr, old_l, new_size) }
+    #[inline] unsafe fn alloc_zeroed(&mut self, l: Layout) -> Result<NonNull<u8>, AllocErr> { self.deref_mut().alloc_zeroed(l) }
+    #[inline] unsafe fn alloc_excess(&mut self, l: Layout) -> Result<Excess, AllocErr> { self.deref_mut().alloc_excess(l) }
+    #[inline] unsafe fn realloc_excess(&mut self, ptr: NonNull<u8>, old_l: Layout, new_size: usize) -> Result<Excess, AllocErr> { self.deref_mut().realloc_excess(ptr, old_l, new_size) }
     #[inline] unsafe fn resize_in_place(&mut self, ptr: NonNull<u8>, old_l: Layout, new_size: usize) -> Result<(), CannotReallocInPlace> { self.deref_mut().resize_in_place(ptr, old_l, new_size) }
 
     #[inline] fn usable_size(&self, l: Layout) -> (usize, usize) { self.deref().usable_size(l) }
